#include <migos_stdio.h>
#include <string.h>
#include <kernel/mm.h>
#include <fs/fat.h>
#include <types.h>

//=============================================================
// init_fatfs
//  fs - fat file system
//  sio - sector i/o functional structure
//-------------------------------------------------------------
// initialize fatfs 
//=============================================================
int init_fatfs( fatfs* fs, sector_io sio )
{
	int8u bootsector[512] = {0,};
	int fat_sectors;
	int8u* cur_fat;
	int s;

	// setup actually sector i/o
	fs->sio = sio;

	// [trick] set default sector size for reading bootsector
	fs->sio.bytes_per_sector = 512;

	if( fs->sio.read == 0 )
		return 0;

	// read bootsector for BPB
	if( fs->sio.read( &fs->sio, 0, 0, bootsector, sizeof(bootsector) ) < sizeof(BPB) )
		return 0;
	raw2BPB( bootsector, &fs->bpb );
	
	fs->type = get_fs_type( &fs->bpb );
	

	// set fast function
	if( fs->type == FS_FAT12 )
	{
		fs->get_next_cluster = get_next_cluster_fat12;
		fs->get_prev_cluster = get_prev_cluster_fat12;
	}
	else
	if( fs->type == FS_FAT16 )
	{
		fs->get_next_cluster = get_next_cluster_fat16;
		fs->get_prev_cluster = get_prev_cluster_fat16;
	}
	else
	{
		fs->get_next_cluster = 0;
		fs->get_prev_cluster = 0;
		return 0;
	}

	// set additional info
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] + (fs->bpb.BPB_BytesPerSec[1] << 8);
//	fs->sio.bytes_per_sector = fs->bpb.BPB_BytesPerSec;
	fs->sio.bytes_per_sector = bytes_per_sec;
	
	// read FAT cache buffer( File Allocation Table ) for performance
	fat_sectors = fs->bpb.BPB_NumFATs * fs->bpb.BPB_FATSz16;

	fs->fat = k_malloc( fat_sectors * bytes_per_sec );

	cur_fat = fs->fat;
	for( s = 1; s < fat_sectors+1; s++ )
	{
		fs->sio.read( &fs->sio, s, 0, cur_fat, bytes_per_sec);
		cur_fat += bytes_per_sec;
	}
	/*
	printf(" fs->fat : 0x%08x fat[25] fat[26] = %02x %02x  = int16u fat[25] = %04x %04x %04x %04x %04x\r\n", fs->fat, fs->fat[25], fs->fat[26], *(short *) ((char*)(fs->fat)+25), *((int16u*)(&fs->fat[25])), *((int16u*)0xA2000019), *((int16u*)0xA2000020), *((int16u*)0xA2000021) );
*/
	return 1;
}

//=============================================================
// free_fatfs
//  fs - fat file system
//-------------------------------------------------------------
// fs's fat cache buffer must be free
//=============================================================
void free_fatfs( fatfs* fs )
{
	if( fs == 0 ) return;
	k_free( fs->fat );
}

//=============================================================
// fatfs_get_fat_count
//  fs - fat file system
//-------------------------------------------------------------
// count fat's bytes
//=============================================================
int fatfs_get_fat_count( fatfs* fs )
{
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] + (fs->bpb.BPB_BytesPerSec[1] << 8);

	if( fs->type == FS_FAT12 )
	{
		return ( fs->bpb.BPB_FATSz16 * bytes_per_sec*8 ) / 12 + 1;
	}
	else
	if( fs->type == FS_FAT16 )
	{
		return ( fs->bpb.BPB_FATSz16 * bytes_per_sec) / 2;
	}
	return 0;
}


//=============================================================
// fatfs_get_fat_count
//  bpb - boot parameter block
//  fat - 
//-------------------------------------------------------------
// count fat's bytes
//=============================================================
int get_cluster_in_fat12( BPB* bpb, char* fat, int i )
{
	/*  Multiply by 1.5 without using floating point, the devide by rounds DOWN */
	int FATOffset = i + (i/2);
	int bytes_per_sec = bpb->BPB_BytesPerSec[0] + (bpb->BPB_BytesPerSec[1] << 8);

	int ThisFATEntOffset = (FATOffset % bytes_per_sec);
	
	int16u FAT12ClusEntryVal;

	if (ThisFATEntOffset == (bytes_per_sec- 1) )
	{
		// 섹터 - 1 바이트 째에서 읽으려고하면 현재 섹터의 마지막 바이트와
		// 다음 섹터의 첫 바이트를 읽어야 하므로 이를 처리하라는 말인데
		// 나는 fat이 메모리에 있으므로 그런 걱정은 없다.
		/* This cluster access spans a sector boundary in the FAT	*/
		/* There are a number of strategies to handling this. The 	*/
		/* easiest is to always load FAT sectors into memory 	*/
		/* in pairs if the volume is FAT12 (if you want to load	*/
		/* FAT sector N, you also load FAT sector N+1 immediately	*/
		/* following it in memory unless sector N is the last FAT	*/
		/* sector). It is assumed that this is the strategy used here	*/
		/* which makes this if test for a sector boundary span	*/
		/* unnecessary.						*/
	}

	FAT12ClusEntryVal = *((int16u *) &fat[ThisFATEntOffset]);
	if (i & 0x0001)
		FAT12ClusEntryVal = FAT12ClusEntryVal >> 4;       /* Cluster number is ODD */
	else
		FAT12ClusEntryVal = FAT12ClusEntryVal & 0x0FFF;/* Cluster number is EVEN */

	return FAT12ClusEntryVal;
}

void set_cluster_in_fat12( BPB* bpb, char* fat, int i, int value )
{
	int FATOffset = i + (i/2);
	int bytes_per_sec = bpb->BPB_BytesPerSec[0] + (bpb->BPB_BytesPerSec[1] << 8);

	int ThisFATEntOffset = (FATOffset % bytes_per_sec);

	if (ThisFATEntOffset == (bytes_per_sec- 1) )
	{
		// 섹터 - 1 바이트 째에서 읽으려고하면 현재 섹터의 마지막 바이트와
		// 다음 섹터의 첫 바이트를 읽어야 하므로 이를 처리하라는 말인데
		// 나는 fat이 메모리에 있으므로 그런 걱정은 없다.
		/* This cluster access spans a sector boundary in the FAT	*/
		/* There are a number of strategies to handling this. The 	*/
		/* easiest is to always load FAT sectors into memory 	*/
		/* in pairs if the volume is FAT12 (if you want to load	*/
		/* FAT sector N, you also load FAT sector N+1 immediately	*/
		/* following it in memory unless sector N is the last FAT	*/
		/* sector). It is assumed that this is the strategy used here	*/
		/* which makes this if test for a sector boundary span	*/
		/* unnecessary.						*/
	}

	if (i & 0x0001)
	{
		*((int16u *) &fat[ThisFATEntOffset]) &= 0x000F;
		*((int16u *) &fat[ThisFATEntOffset]) |= ((int16u)value << 4);
	}
	else
	{
		*((int16u *) &fat[ThisFATEntOffset]) &= 0xF000;
		*((int16u *) &fat[ThisFATEntOffset]) |= ((int16u)value & 0x0FFF );
	}
}


int get_next_cluster_fat12( fatfs* fs, int cluster )
{
	return get_cluster_in_fat12( &fs->bpb, fs->fat, cluster );
}

int get_prev_cluster_fat12( fatfs* fs, int cluster )
{
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] + (fs->bpb.BPB_BytesPerSec[1] << 8);
	// FAT Table (Cluster Allocation Table)
	int fat_count = ( fs->bpb.BPB_FATSz16 * bytes_per_sec*8 ) / 12 + 1;
	int i;

	for( i = 0 ; i < fat_count ; i ++ )
	{
		if( cluster == get_cluster_in_fat12( &fs->bpb, fs->fat, i ) )
			return i;
	}
	// not found	
	return NOT_FOUND_CLUSTER;
}

int get_next_cluster_fat16( fatfs* fs, int cluster )
{
	// FAT Table (Cluster Allocation Table)
	int8u* fat = (int8u*)fs->fat;
	
	// for FAT16
	int16u* fat16 = (int16u*)fat;

	// get next cluster
	return fat16[ cluster ];
}

int get_prev_cluster_fat16( fatfs* fs, int cluster )
{
	// FAT Table (Cluster Allocation Table)
	int8u* fat = (int8u*)fs->fat;
	// for FAT16
	int16u* fat16 = (int16u*)fat;
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] + (fs->bpb.BPB_BytesPerSec[1] << 8);

	int fat_count = ( fs->bpb.BPB_FATSz16 * bytes_per_sec) / 2;
	int i;
	for( i = 0 ; i < fat_count ; i ++ )
	{
		if( fat16[i] == cluster )
			return i;
	}
	// not found	
	return NOT_FOUND_CLUSTER;
}

int fatfs_cluster_set_directory_entry( fatfs* fs, fatfs_path* wc )
{
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] + (fs->bpb.BPB_BytesPerSec[1] << 8);

	if( wc == 0 || wc->de == 0 )
		return 0;

	// set directory entry
	while( 1 ) 
	{
		char* limit;

		// root cluster isn't equal general cluster limit
		if( wc->cluster == ROOT_CLUSTER && (fs->type == FS_FAT12 || fs->type == FS_FAT16) )
			limit = (wc->buf+FATFS_MAX_PATH_BUF_SECTOR*bytes_per_sec);
		else
			limit = (wc->buf+fs->bpb.BPB_SecPerClus*bytes_per_sec);

		// over cluster buffer then auto next cluster
		if( limit <= (char*)(wc->de+1) )
		{
			// end of cluster?
			if( fatfs_read_next_cluster( fs, wc ) == END_OF_CLUSTER )
				return 0;
		}

		if( wc->de->DIR_FileName[0] != 0xE5 ) // unused
			break;

		if( wc->de->DIR_FileName[0] == 0x00 ) // end of cluster
			return 0;

		wc->de++;
	}

	if( wc->de->DIR_FileName[0] == 0x00 ) // end of cluster
		return 0;

	fatfs_get_file_full_name( wc->FileName, wc->de->DIR_FileName, wc->de->DIR_Extension );
	wc->Attr = wc->de->DIR_Attr;
	wc->CrtDate = wc->de->DIR_CrtDate;
	wc->CrtTime = wc->de->DIR_CrtTime;
	wc->WrtDate = wc->de->DIR_WrtDate;
	wc->WrtTime = wc->de->DIR_WrtTime;
	wc->LstAccDate = wc->de->DIR_LstAccDate;
	wc->FileSize = wc->de->DIR_FileSize;

	return 1;
}

int fatfs_read_cluster( fatfs* fs, fatfs_path* wc, int sector, int sector_count )
{
	int8u* buf;
	int i, s;
	
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] + (fs->bpb.BPB_BytesPerSec[1] << 8);
	
	s = sector;
	buf = wc->buf;

	for( i = 0; i < sector_count; i ++ )
	{
		// copy to work buffer for cluster
		fs->sio.read( &fs->sio, s, 0, buf, bytes_per_sec);
		
		buf += bytes_per_sec; // cluster work buffer next
		s++;	// next root sector
	}

	// set directory entry
	wc->de = (directory_entry*)wc->buf;

	fatfs_get_file_full_name( wc->FileName, wc->de->DIR_FileName, wc->de->DIR_Extension );

	//fatfs_cluster_set_directory_entry( fs, wc );
/*
	if( wc->cluster )	// root is cluster=0 then no first auto scan
	{
		fatfs_cluster_set_directory_entry( fs, wc );
	}
	else
	{
		fatfs_get_file_full_name( wc->FileName, wc->de->DIR_FileName, wc->de->DIR_Extension );
		wc->Attr = wc->de->DIR_Attr;
		wc->CrtDate = wc->de->DIR_CrtDate;
		wc->CrtTime = wc->de->DIR_CrtTime;
		wc->WrtDate = wc->de->DIR_WrtDate;
		wc->WrtTime = wc->de->DIR_WrtTime;
		wc->LstAccDate = wc->de->DIR_LstAccDate;
		wc->FileSize = wc->de->DIR_FileSize;
	}
*/
	return i;
}

int fatfs_read_work_cluster( fatfs* fs, fatfs_path* wc, int cluster )
{
	int start_sector;
	start_sector = get_first_sector_of_cluster( &fs->bpb, cluster );
	
	wc->cluster = cluster;
	return fatfs_read_cluster( fs, wc, start_sector, fs->bpb.BPB_SecPerClus );
}

int fatfs_read_root_cluster( fatfs* fs, fatfs_path* wc )
{
	int start_sector, sector_count;
	start_sector = get_root_dir_sector( &fs->bpb );
	sector_count = get_root_dir_sectors( &fs->bpb );
	
	// todo: path's cluster buffer size
	// root sectors = 32 sectors..
	// but buffer is FATFS_MAX_PATH_BUF_SECTOR..
	if( sector_count > FATFS_MAX_PATH_BUF_SECTOR )
		sector_count = FATFS_MAX_PATH_BUF_SECTOR;

	// todo: root cluster ? = 0
	wc->cluster = ROOT_CLUSTER;
	wc->cur_dir_cluster = wc->cluster;

	return fatfs_read_cluster( fs, wc, start_sector, sector_count );
}

int fatfs_get_next_cluster( fatfs* fs, int cluster )
{
	int next_cluster = fs->get_next_cluster( fs, cluster );
	
	// next_cluster is end of cluster?
	if( fs->type == FS_FAT16 )
	{
		if( (int16u)(next_cluster&0xFFF0) == 0xFFF0 ) 
			next_cluster = END_OF_CLUSTER;
	}
	else
	if( fs->type == FS_FAT12 )
	{
		if( (int16u)(next_cluster&0x0FFF) == 0x0FFF ) 
			next_cluster = END_OF_CLUSTER;
	}
	//else
		// unsupported type
		//return 0;

	return next_cluster;
}

int fatfs_get_last_cluster( fatfs* fs, int cluster )
{
	int next_cluster;
	while( 1 )
	{
		next_cluster = fatfs_get_next_cluster( fs, cluster );
		if( next_cluster == END_OF_CLUSTER )
			return cluster;
		cluster = next_cluster;
	}

	return NOT_FOUND_CLUSTER;
}

int fatfs_get_first_cluster( fatfs* fs, int cluster )
{
	int prev_cluster;
	while( 1 )
	{
		prev_cluster = fatfs_get_prev_cluster( fs, cluster );
		if( prev_cluster == NOT_FOUND_CLUSTER )
			return cluster;
		cluster = prev_cluster;
	}

	return NOT_FOUND_CLUSTER;
}

int fatfs_read_next_cluster( fatfs* fs, fatfs_path* wc )
{
	int next_cluster = fatfs_get_next_cluster( fs, wc->cluster );
	if( next_cluster != END_OF_CLUSTER )
		fatfs_read_work_cluster( fs, wc, next_cluster );
	return next_cluster;
}

int fatfs_get_prev_cluster( fatfs* fs, int cluster )
{
	int prev_cluster = 0;

	prev_cluster = fs->get_prev_cluster( fs, cluster );

	return prev_cluster;
}

int fatfs_read_prev_cluster( fatfs* fs, fatfs_path* wc )
{
	int prev_cluster;
	prev_cluster = fs->get_prev_cluster( fs, wc->cluster );

	if( prev_cluster != 0 )
		fatfs_read_work_cluster( fs, wc, prev_cluster );
	return prev_cluster;
}

// ex) sub_dir , hello.txt, sub_dir/hello.txt
fatfs_path* fatfs_path_recursive( fatfs* fs, fatfs_path* p, char* path )
{
	char* token_end, *token_end1, *token_end2;
	char token[MAX_FSPATH];

	// error or not found dir/file
	if( p->de == 0 )
		return p;

	// end of path
	if( path == 0 || *path == '\0' || strcmp(path,"/") == 0 || strcmp(path,"\\") == 0 )
	{
		return p;
	}

	// /sub => sub
	if( *path == '/' || *path == '\\' )
		path++;

	// get this-time token of path
	token_end1 = strchr( path, '/' );
	token_end2 = strchr( path, '\\' );
	
	token_end = token_end1;
	if( token_end1 < token_end2 )
		token_end = token_end2;
	
	// make current dir/file token
	if( token_end != 0 )
	{
		// directory 
		strncpy( token, path, token_end - path );
		token[token_end - path] = '\0';	
	}
	else
	{
		// terminal node file
		strcpy( token, path );
	}

	// find dir/file in this clusters
	while( p && p->de )
	{
		if( fatfs_is_equal_fname( p->de->DIR_FileName, token ) )
			break;
		
		// next directory entry
		p->de++;
		// fix available directroy entry
		if( fatfs_cluster_set_directory_entry( fs, p ) == 0 )
		{
			p->de = 0;
			return p;
		}
	}

	if( p == 0 || p->de == 0 )
	{
		// not found in this clusters.
		return p;
	}

	if( p->de->DIR_Attr & FILE_ATTR_DIRECTORY )
	{
		int sub_cluster;
		sub_cluster = (int)p->de->DIR_FstClusLO;
		fatfs_read_work_cluster( fs, p, sub_cluster );
		
		p->cur_dir_cluster = sub_cluster;

		if( sub_cluster == ROOT_CLUSTER )
		{
			p->cur_dir_cluster = ROOT_CLUSTER;
			p->cluster = ROOT_CLUSTER;
			strcpy( p->FullPath, "/" );
			strcpy( p->FileName, "/" );
			p->Attr = FILE_ATTR_DIRECTORY|FILE_ATTR_SYSTEM|FILE_ATTR_DIRECTORY;	
			return p;
		}
	}

	return fatfs_path_recursive( fs, p, token_end );
}


// ex) sub_dir , hello.txt, sub_dir/hello.txt
//fatfs_path* fatfs_open_path( fatfs* fs, char* path )
int fatfs_open_path( fatfs* fs, char* path, fatfs_path* p )
{
	//char fullpath[MAX_FSPATH*2];
	
	//memset( fullpath, 0, sizeof(fullpath) );
	if( p->de != 0 && *path != '/' && strcmp( p->FullPath, "/") != 0 )
	{
		//strcpy( fullpath, p->FileName );
		//strcat( fullpath, "/" );
		//strcat( fullpath, path );

		fatfs_read_work_cluster( fs, p, p->cluster );
	}
	else
	{
		if( strcmp( p->FullPath, "/" ) == 0 && ( strcmp( path, "." ) ==0 || strcmp( path, ".." ) == 0 ) )
			return 1;

		//strcpy( fullpath, path );

		// can't get root cluster
		if( fatfs_read_root_cluster( fs, p ) == 0 )
			return 0;
	}


	// is just root ?
	if( path == 0 || *path == 0 || strcmp(path,"/") == 0 || strcmp(path,"\\") == 0 )
	{
		p->cur_dir_cluster = ROOT_CLUSTER;
		p->cluster = ROOT_CLUSTER;
		strcpy( p->FullPath, "/" );
		strcpy( p->FileName, "/" );
		p->Attr = FILE_ATTR_DIRECTORY|FILE_ATTR_SYSTEM|FILE_ATTR_DIRECTORY;
		return 1;
	}

	// else.. now travel our path..
	if( *path == '/' || *path == '\\' )
		path++;

	p = fatfs_path_recursive( fs, p, path );

	if( p->de )
	{
		fatfs_get_full_path( fs, p, p->FullPath, sizeof(p->FullPath) );
		return 1;
	}

	return 0;
}

char* get_path_token( char* path )
{
	char* token_end1, *token_end2, *token_end;
	
	// get this-time token of path
	token_end1 = strchr( path, '/' );
	token_end2 = strchr( path, '\\' );
	
	token_end = token_end1;
	if( token_end1 < token_end2 )
		token_end = token_end2;
	return token_end;
}

char* get_path_prev( char* cur, char* start )
{
	while( cur >= start )
	{
		if( *cur == '/' || *cur == '\\' )
		{
			cur++;
			break;
		}
		cur--;
	}

	// not found
	if( cur < start )
		cur = 0;
	return cur;
}


fatfs_path* fatfs_path_next_internal( fatfs* fs, fatfs_path* wc )
{
	//if( strcmp( wc->FullPath , "/" ) != 0 )
	// next directory entry
	wc->de++;

	// fix available directroy entry
	if( fatfs_cluster_set_directory_entry( fs, wc ) == 0 )
	{
		wc->de = 0;
		return wc;
	}
	return wc;
}

fatfs_path* fatfs_path_next( fatfs* fs, fatfs_path* wc )
{
	if( wc->de == 0 )
		return wc;

	fatfs_path_next_internal( fs, wc );

	if( wc->de )
		fatfs_get_full_path( fs, wc, wc->FullPath, sizeof(wc->FullPath) );
	return wc;
}

char* fatfs_get_last_char( char* psz, int len )
{
	char* cur = psz+len-1;
	while( cur != psz )
	{
		if( *cur != ' ' )
			break;
		cur--;
	}
	if( cur == psz && *cur != *psz )
		return 0;
	return cur;
}


void fatfs_get_file_full_name( char* fullname, char* filename, char* extension )
{
	char* last;
	char* cur;
	
	// process file name
	last = fatfs_get_last_char( filename, 8 );
	cur = filename;
	while( cur <= last )
		*fullname++ = *cur++;
	
	// process extension
	last = fatfs_get_last_char( extension, 3 );
	if( last != 0 && *extension != ' ')
	{
		*fullname++ = '.';
		cur = extension;
		while( cur <= last )
			*fullname++ = *cur++;
	}
	*fullname = '\0';
}

// support 8.3 FAT12/16 short filename
int fatfs_is_equal_fname( char* DIR_FName, char* filename )
{
	int i;
	char* cur_DIR, *cur_fname;

	cur_DIR = DIR_FName;
	cur_fname = filename;

	// file name part
	for( i = 0 ; i < 8 ; i ++ )
	{
		if( toupper( DIR_FName[i] ) == toupper( *cur_fname ) )
			cur_fname++;
		else
		if( DIR_FName[i] == ' ' && ( *cur_fname == '.' || *cur_fname == '\0' ) )
			continue;
		else
			return 0;
	}
	
	if( *cur_fname != '\0' )
	{
		if( *cur_fname != '.' || *(cur_fname+1) == '\0' )
			return 0;

		cur_fname++;		// skip '.'
	}
	
	// extension part
	for( i = 8 ; i < 11 ; i ++ )
	{
		if( DIR_FName[i] == ' ' && *cur_fname == 0 )
			continue;
		
		if( toupper( DIR_FName[i] ) != toupper( *cur_fname ) )
			return 0;
		
		cur_fname++;
	}

	return 1;
}

int fatfs_is_directory( fatfs_path* wc )
{
	if( wc == 0 || wc->de == 0 )
		return 0;
	return (wc->de->DIR_Attr & FILE_ATTR_DIRECTORY)!=0;
}

int fatfs_fix_full_path( char* path, char* fixbuf, int size )
{
	// /sub/some/.././../temp => /temp
	char* cur = path;
	char* fix = fixbuf;
	char* cur_end;

	memset( fixbuf, 0, size );

	while( cur && *cur && fix < (fixbuf+size) )
	{
		// dir
		if( *cur == '/' || *cur == '\\' )
		{
			if( fix == fixbuf || (fix > fixbuf && *(fix-1) != '/' && *(fix-1) != '\\') )
			{
				*fix = *cur;
				fix++;
			}
			cur++;
		}

		cur_end = get_path_token( cur );
		if( !cur_end )
		{
			//terminal
			strcpy( fix, cur );
			break;
		}
		else
		{
			char* cur_dir = cur;
			
			cur = cur_end;
			
			// omit dot directories
			if( strncmp( cur_dir, "..", 2 ) == 0 )
			{
				// back to parent
				char* parent = get_path_prev( fix-2, fixbuf );
				fix = parent;
				if( parent == 0 )
					fix = fixbuf;
			}
			else
			if( strncmp( cur_dir, ".", 1 ) == 0 )
			{
				// nothing 
			}
			else
			{
				int len = cur_end - cur_dir;
				strncpy( fix, cur_dir, len );
				fix += len;
			}
		}
	}
	
	return 1;
}

int fatfs_read_parent_cluster( fatfs* fs, int cur_dir_cluster, fatfs_path* p )
{
	fatfs_path dir;
	
	// trick cluster == 1 ( root )
	if( cur_dir_cluster == ROOT_CLUSTER )
		return 0;
	
	fatfs_read_work_cluster( fs, &dir, cur_dir_cluster );
	
	while( dir.de != 0 )
	{
		if( strncmp( dir.de->DIR_FileName, "..", 2 ) == 0 )
		{
			fatfs_read_work_cluster( fs, p, dir.de->DIR_FstClusLO );

			while( p->de != 0 )
			{
				if( p->de->DIR_FstClusLO == cur_dir_cluster )
				{
					//fatfs_read_work_cluster( fs, &dir, dir.de->DIR_FstClusLO );

					return 1;
				}
				fatfs_path_next_internal( fs, p );
			}

			// not found
			return 0;
		}

		fatfs_path_next_internal( fs, &dir );
	}
	
	return 0;	
}

int fatfs_get_full_path( fatfs* fs, fatfs_path* p, char* buf, int len )
{
	fatfs_path dir;

	int dir_cluster = p->cur_dir_cluster;
	char* cur = buf+len-1;
	int curlen;

	memset( buf, 0, len );

	if( strcmp( p->FullPath , "/" ) == 0 || 
		strcmp( p->FileName , "/" ) == 0 )
	{
		strcpy( buf, "/" );
		return 1;
	}

	if( fatfs_is_directory( p ) && strncmp( p->FileName, ".", 1 ) != 0 )
		*cur-- = '/';
	
	if( strcmp( p->FileName, "." ) != 0 )
	{
		curlen = strlen( p->FileName );
		cur -= (curlen-1);
		memcpy( cur, p->FileName, curlen );
	}

	while( fatfs_read_parent_cluster( fs, dir_cluster, &dir ) )
	{
		curlen = strlen( dir.FileName );
		cur -= (curlen+1);
		memcpy( cur, dir.FileName, curlen );
		cur[curlen] = '/';
		
		dir_cluster = dir.cluster;
	}

	// set root dir
	*(cur-1) = '/';
	memmove( buf, cur-1, len - (cur-1-buf) );

	return 1;
}




//////////////////////////////////////////////////////////////////////
// 부트섹터 메모리블럭에서 BPB 정보를 뽑는다.
// FAT12 FAT16 FAT32
//////////////////////////////////////////////////////////////////////
void raw2BPB( char* data, BPB* bpb )
{
	// little-endian 인 경우로 가정한다.
	// FAT은 intel과 호환하려면 내부 데이타가 little-endian이여야 한다.
	memcpy( bpb, data, sizeof(BPB) );
}

int is_bootsector( char* bootsector )
{
	return (bootsector[510] == 0x55 && bootsector[511] == 0xAA )?1:0;
}

//////////////////////////////////////////////////////////////////////
// 루트디렉토리의 시작 섹터
// FAT12 FAT16
//////////////////////////////////////////////////////////////////////
int get_root_dir_sector( BPB* bpb )
{
	// FAT32는 루트 디렉토리가 고정이 아니라 클러스터로 구성된 체인 안에 있다.
	// FAT12/16 경우만 지원
	int FirstRootDirSecNum = bpb->BPB_RsvdSecCnt + (bpb->BPB_NumFATs * bpb->BPB_FATSz16);
	return FirstRootDirSecNum;
}

//////////////////////////////////////////////////////////////////////
// 루트 디렉토리가 사용하는 섹터의 수
// FAT12 FAT16
//////////////////////////////////////////////////////////////////////
int get_root_dir_sectors( BPB* bpb )
{
	int root_ent_cnt = bpb->BPB_RootEntCnt[0] +(bpb->BPB_RootEntCnt[1] << 8);
	int bytes_per_sec = bpb->BPB_BytesPerSec[0] +(bpb->BPB_BytesPerSec[1] << 8);

	int RootDirSectors = ((root_ent_cnt*32) + (bytes_per_sec-1)) / bytes_per_sec;

	return RootDirSectors;

}

//////////////////////////////////////////////////////////////////////
// 데이타 영역의 첫 섹터
// FAT12 FAT16
//////////////////////////////////////////////////////////////////////
int get_first_data_sector( BPB* bpb )
{
	/*
		If (BPB_FATSz16 != 0)
		FATSz = BPB_FATSz16;
	Else
		FATSz = BPB_FATSz32;
		FirstDataSector = BPB_RsvedSecCnt + (BPB_NumFATs * FATSz) + RootDirSectors;
	*/
	int FirstDataSector = bpb->BPB_RsvdSecCnt + (bpb->BPB_NumFATs * bpb->BPB_FATSz16) + get_root_dir_sectors( bpb );
	return FirstDataSector;
}

// 해당 클러스터가 할당된 영역의 첫 섹터
// 해당 클러스터가 파일 데이타면 실제 파일의 내용이 들어있다.
// 해당 클러스터가 디렉토리 라면 디렉토리 엔트리 배열이 들어 있다. 
int get_first_sector_of_cluster( BPB* bpb, int cluster )
{
	int FirstSectorofCluster;

	// cluster 0번을 요구하면 root 의 첫 섹터를 준다. ( by bro )
	if( cluster == ROOT_CLUSTER )
		return get_root_dir_sector( bpb );

	FirstSectorofCluster = ((cluster - 2) * bpb->BPB_SecPerClus) + get_first_data_sector(bpb);
	return FirstSectorofCluster;
}

// 데이타 영역이 차지하는 총 섹터 수
int get_data_sectors( BPB* bpb )
{
	/*
	If (BPB_FATSz16 != 0)
		FATSz = BPB_FATSz16;
	Else
		FATSz = BPB_FATSz32;

	If (BPB_TotSec16 != 0)
		TotSec = BPB_TotSec16;
	Else
		TotSec = BPB_TotSec32;	
	*/


	
	kernel_printf("BPB size : %d \n", sizeof(struct _BPB));
	kernel_printf("Directory entry size : %d \n", sizeof(struct _directory_entry));
	int DataSec = bpb->BPB_TotSec16 - (bpb->BPB_RsvdSecCnt + (bpb->BPB_NumFATs * bpb->BPB_FATSz16) + get_root_dir_sectors(bpb));
	kernel_printf("finish bpb \n");
	return DataSec;
}


int get_find_cluster_in_fat12( BPB* bpb, int8u* fat, int cluster )
{
	int bytes_per_sec = bpb->BPB_BytesPerSec[0] +(bpb->BPB_BytesPerSec[1] << 8);

	// for FAT12
	int fat_count = ( bpb->BPB_FATSz16 * bytes_per_sec*8 ) / 12 + 1;
	int i;

	for( i = 0 ; i < fat_count ; i ++ )
	{
		if( cluster == get_cluster_in_fat12( bpb, fat, i ) )
			return i;
	}
	// not found	
	return NOT_FOUND_CLUSTER;
}

int get_find_cluster_in_fat16( BPB* bpb, int8u* fat, int cluster )
{
	// for FAT16
	int16u* fat16 = (int16u*)fat;
	int bytes_per_sec = bpb->BPB_BytesPerSec[0] +(bpb->BPB_BytesPerSec[1] << 8);

	int fat_count = ( bpb->BPB_FATSz16 * bytes_per_sec) / 2;
	int i;
	for( i = 0 ; i < fat_count ; i ++ )
	{
		if( fat16[i] == cluster )
			return i;
	}
	// not found	
	return NOT_FOUND_CLUSTER;
}

// 총 클러스터의 수
int get_clusters( BPB* bpb )
{
	//CountofClusters = DataSec / BPB_SecPerClus; 계산후 소수점 내림
	int CountofClusters = get_data_sectors(bpb) / bpb->BPB_SecPerClus;
	return CountofClusters;
}

fs_type get_fs_type( BPB* bpb )
{
	int clusters = get_clusters( bpb );
	if( clusters < 4085 )
		return FS_FAT12;
	else
	if( clusters < 65525 )
		return FS_FAT16;
	return FS_FAT32;
}

// 디렉토리 파일명이 0 또는 0xE5이면 지워진 파일이거나 쓰이지 않은 파일이다. 
// 특히 0이면 이 후의 디렉토리 엔트리는 할당된것이 없음을 알린다. 디렉토리 끝
// 0x05는 일본어용으로 실제로는 0xE5이다. (이미 unused용으로 0xE5가 사용되어서)
int is_unused_entry( directory_entry* pde )
{
	if( *(char*)pde == 0xE5 || *(char*)pde == 0x0 )
		return 1;
	else
		return 0;
}






int fatfs_fopen( fatfs* fs, char* path, fatfs_file* file )
{
	fatfs_path p;
	if( fatfs_open_path( fs, path, &p ) == 0 )
		return 0;
	if( fatfs_is_directory( &p ) )
		return 0;

	// file i/o
	fatfs_get_file_full_name( file->FileName, p.de->DIR_FileName, p.de->DIR_Extension );
	strcpy( file->FullPath , p.FullPath );
	file->Attr = p.de->DIR_Attr;
	file->CrtDate = p.de->DIR_CrtDate;
	file->CrtTime = p.de->DIR_CrtTime;
	file->WrtDate = p.de->DIR_WrtDate;
	file->WrtTime = p.de->DIR_WrtTime;
	file->LstAccDate = p.de->DIR_LstAccDate;
	file->FileSize = p.de->DIR_FileSize;
	file->first_cluster = p.de->DIR_FstClusLO;
	file->cur_sector = 0;
	file->cur_offset = 0;
	file->cur_cluster = file->first_cluster;

	return 1;
}

void fatfs_fclose( fatfs* fs, fatfs_file* file )
{
}

int fatfs_fseek( fatfs* fs, fatfs_file* file, int offset, int orgin )
{
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);

	if( orgin == FATFS_FSEEK_SET )
	{
		file->cur_cluster = file->first_cluster;
		file->cur_sector = 0;
		file->cur_offset = 0;
		if( offset < 0 ) offset=-offset;

		return fatfs_fseek( fs, file, offset, FATFS_FSEEK_CUR );
	}
	else
	if( orgin == FATFS_FSEEK_END )
	{
		int end_cluster = 0;
		int cluster = file->first_cluster;
		while( cluster != END_OF_CLUSTER )
		{
			end_cluster = cluster;
			cluster = fatfs_get_next_cluster( fs, cluster );
		}

		file->cur_cluster = end_cluster;
		file->cur_sector = fs->bpb.BPB_SecPerClus-1;
		file->cur_offset = bytes_per_sec;
		if( offset < 0 ) offset=-offset;

		return fatfs_fseek( fs, file, -offset, FATFS_FSEEK_CUR );
	}
	else
	if( orgin == FATFS_FSEEK_CUR )
	{
		int cur_cluster = file->cur_cluster;
		int cur_sector = file->cur_sector;
		int cur_offset = file->cur_offset;

		if( offset > 0 )
		{
			if( cur_offset+offset < bytes_per_sec)
			{
				cur_offset += offset;
			}
			else
			{
				int amount = 0;
				int left_offset = offset - (bytes_per_sec-cur_offset);
				while( left_offset > 0 )
				{
					cur_offset=0;
					cur_sector++;

					if( cur_sector >= fs->bpb.BPB_SecPerClus )
					{
						int next_cluster;
						next_cluster = fatfs_get_next_cluster( fs, cur_cluster );
						if( next_cluster == END_OF_CLUSTER )
						{
							// EOF
							cur_sector = fs->bpb.BPB_SecPerClus-1;
							cur_offset = bytes_per_sec-1;
							break;
						}
						cur_cluster = next_cluster;
						cur_sector = 0;
					}

					amount = left_offset;
					if( amount > (int)(bytes_per_sec-1) )
						amount = (int)(bytes_per_sec-1);

					left_offset -= amount;
					cur_offset -= amount;
				}
			}
		}
		else
		if( offset < 0 )
		{
			offset=-offset;

			if( cur_offset > (int)offset )
			{
				cur_offset -= offset;
			}
			else
			{
				int amount = 0;
				int left_offset = offset - cur_offset;
				while( left_offset > 0 )
				{
					cur_offset=bytes_per_sec;
					cur_sector--;

					if( cur_sector < 0 )
					{
						int prev_cluster;
						prev_cluster = fatfs_get_prev_cluster( fs, cur_cluster );
						if( prev_cluster == NOT_FOUND_CLUSTER )
						{
							// BOF
							cur_sector = 0;
							cur_offset = 0;
							break;
						}
						cur_cluster = prev_cluster;
						cur_sector = fs->bpb.BPB_SecPerClus-1;
					}

					amount = left_offset;
					if( amount > bytes_per_sec)
						amount = bytes_per_sec;

					left_offset -= amount;
					cur_offset -= amount;
				}
			}
		}

		file->cur_cluster = cur_cluster;
		file->cur_sector = cur_sector;
		file->cur_offset = cur_offset;
	}

	return 0;
}

int fatfs_ftell( fatfs* fs, fatfs_file* file )
{
	int clusters = 0;
	int cluster = file->first_cluster;
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);

	while( cluster != END_OF_CLUSTER && cluster != file->cur_cluster )
	{
		cluster = fatfs_get_next_cluster( fs, cluster );
		clusters++;
	}

	return clusters * fs->bpb.BPB_SecPerClus * bytes_per_sec+
				file->cur_sector * bytes_per_sec+
				file->cur_offset;
}

int fatfs_fread( fatfs* fs, fatfs_file* file, void* buf, int size )
{
	int cur_cluster;
	int cur_sector;
	int cur_offset;
	int left = size;
	int8u* cur = (int8u*)buf;
	int total_readed = 0;
	
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);
	
	int left_of_file;
	left_of_file = (int)file->FileSize - fatfs_ftell( fs, file );
	if( left > left_of_file )
		left = left_of_file;
	
	cur_cluster = file->cur_cluster;
	cur_sector = file->cur_sector;
	cur_offset = file->cur_offset;

	while( 1 )
	{
		int data_sector, readed;

		// check boundary of sector
		if( cur_offset >= bytes_per_sec)
		{
			cur_sector++;
			cur_offset = 0;
		}

		// check boundary of cluster
		if( cur_sector >= fs->bpb.BPB_SecPerClus )
		{
			int next_cluster = 0;
			next_cluster = fatfs_get_next_cluster( fs, cur_cluster );
			if( next_cluster == END_OF_CLUSTER )
				break;

			cur_cluster = next_cluster;
			cur_sector = 0;
			cur_offset = 0;
		}

		// 'end of read' or 'end of file'
		if( left <= 0 )
			break;

		data_sector = get_first_sector_of_cluster( &fs->bpb, cur_cluster ) + cur_sector;
		
		readed = fs->sio.read( &fs->sio, data_sector, cur_offset , cur, left );
		//readed = fatfs_read_sector( fs, data_sector, cur_offset , cur, left );
		cur_offset += cur_offset+readed;

		cur += readed;
		left -= readed;
		
		total_readed += readed;
	}

	file->cur_cluster = cur_cluster;
	file->cur_sector = cur_sector;
	file->cur_offset = cur_offset;

	return total_readed;
}

// one sector - test ok 
// over sectors - test ok
int fatfs_fwrite( fatfs* fs, fatfs_file* file, void* buf, int size )
{
	int cur_cluster;
	int cur_sector;
	int cur_offset;
	int left = size;
	int8u* cur = (int8u*)buf;
	int total_written = 0;
	
	int cur_total_pos = fatfs_ftell( fs, file );
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);

	left = size;
	
	cur_cluster = file->cur_cluster;
	cur_sector = file->cur_sector;
	cur_offset = file->cur_offset;

	while( 1 )
	{
		int data_sector, written, write;

		// prepare sector for writing
		// check boundary of sector
		if( cur_offset >= bytes_per_sec)
		{
			cur_sector++;
			cur_offset = 0;
		}

		// check boundary of cluster
		if( cur_sector >= fs->bpb.BPB_SecPerClus )
		{
			int next_cluster = 0;
			
			// if EOF then alloc new file cluster
			next_cluster = fatfs_get_next_cluster( fs, cur_cluster );
			if( next_cluster == END_OF_CLUSTER )
			{
				// alloc new cluster
				next_cluster = fatfs_realloc_cluster( fs, cur_cluster );
			}

			//next_cluster = fatfs_get_next_cluster( fs, cur_cluster );
			if( next_cluster == END_OF_CLUSTER )
				break;

			cur_cluster = next_cluster;
			cur_sector = 0;
			cur_offset = 0;
		}

		// 'end of read' or 'end of file'
		if( left <= 0 )
			break;

		// write on sector
		write = left;
		if( bytes_per_sec< write + cur_offset )
			write = bytes_per_sec- cur_offset;

		data_sector = get_first_sector_of_cluster( &fs->bpb, cur_cluster ) + cur_sector;
		written = fs->sio.write( &fs->sio, data_sector, cur_offset , cur, left );
		cur_offset += written;

		cur += written;
		left -= written;
		
		total_written += written;
	}

	file->cur_cluster = cur_cluster;
	file->cur_sector = cur_sector;
	file->cur_offset = cur_offset;

	// file info update ( in directory entry )
	{
	int32u dwCurFileSize;
	dwCurFileSize = cur_total_pos + total_written;

	if( dwCurFileSize > file->FileSize )
	{
		int tot_offset,sectors, offset, data_sector;
		fatfs_path p;
		
		file->FileSize = dwCurFileSize;

		if( fatfs_open_path( fs, file->FullPath, &p ) == 0 )
			return total_written;
		
		data_sector = get_first_sector_of_cluster( &fs->bpb, p.cluster );
		tot_offset = (char*)p.de - (char*)p.buf;
		sectors = tot_offset / bytes_per_sec;
		offset = tot_offset % bytes_per_sec;
		data_sector += sectors;

		offset = offset + sizeof(directory_entry) - 4;
		fs->sio.write( &fs->sio, data_sector, offset , (char*)&dwCurFileSize, 4 );
	}
	}

	return total_written;
}

void fatfs_set_file_full_name( char* DIR_FileName, char* filefullname )
{
	char* src,*des;
	int i;
	
	src = DIR_FileName;
	des = filefullname;

	// process file name
	for( i = 0 ; i < 8; i ++ )
	{
		if( *des == '\0' || *des == '.' )
		{
			*src++ = ' ';
			continue;
		}
		*src++ = toupper(*des++);
	}
	// skip dot
	if( *des == '.' )
		des++;
	// process extension
	for( i = 0 ; i < 3; i ++ )
	{
		if( *des == '\0' )
		{
			*src++ = ' ';
			continue;
		}
		*src++ = toupper(*des++);
	}
}

char* get_parent_path( char* path, char* parent )
{
	char* parent_end, *path_end;
	int len;

	len = strlen( path );
	if( path[len-1] == '/' || path[len-1] == '\\' )
		path_end = &path[len-2];
	else
		path_end = &path[len-1];

	parent_end = get_path_prev( path_end, path );
	if( parent_end )
	{
		strncpy( parent, path, parent_end - path );
		parent[ parent_end - path ] = '\0';
	}
	else
		strcpy( parent, path );

	return parent_end == 0 ? 0 : parent_end;
}

int fatfs_find_free_cluster( fatfs* fs )
{
	if( fs->type == FS_FAT12 )
		return get_find_cluster_in_fat12( &fs->bpb, fs->fat, 0 );
	else
	if( fs->type == FS_FAT16 )
		return get_find_cluster_in_fat16( &fs->bpb, fs->fat, 0 );
	return NOT_FOUND_CLUSTER;
}

int fatfs_free_cluster_list( fatfs* fs, int cluster )
{
	int next_cluster, del_cluster;
	int first_cluster = fatfs_get_first_cluster( fs, cluster );
	if( first_cluster == NOT_FOUND_CLUSTER )
		return 0;
	
	del_cluster = first_cluster;
	while( 1 )
	{
		next_cluster = fatfs_get_next_cluster( fs, cluster );
		if( next_cluster == END_OF_CLUSTER )
		{
			fatfs_set_fat( fs, cluster, EMPTY_CLUSTER );
			break;
		}
		fatfs_set_fat( fs, cluster, EMPTY_CLUSTER );

		if( cluster == next_cluster ) // error 
			return 0;

		cluster = next_cluster;
	}

	return 1;
}

int fatfs_get_fat( fatfs* fs, int cluster )
{
	if( fs->type == FS_FAT12 )
	{
		return get_cluster_in_fat12( &fs->bpb, fs->fat, cluster );
	}
	else
	if( fs->type == FS_FAT16 )
	{
		int16u* fat16 = (int16u*)fs->fat;
		return fat16[cluster];
	}
	return 0;
}

void fatfs_set_fat( fatfs* fs, int cluster, int value )
{
	BPB* bpb = &fs->bpb;
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);

	if( fs->type == FS_FAT12 )
	{
		int FATOffset = cluster + (cluster/2);

		int ThisFATSecNum = bpb->BPB_RsvdSecCnt + ( FATOffset / bytes_per_sec);
		int ThisFATEntOffset = (FATOffset % bytes_per_sec);

		// update fat memory
		set_cluster_in_fat12( bpb, fs->fat, cluster, value );

		// update fat real sector
		if (ThisFATEntOffset == (bytes_per_sec- 1) )
		{
			// two bytes is on 2 sectors
			// untested
			// 섹터 - 1 바이트 째에서 읽으려고하면 현재 섹터의 마지막 바이트와
			// 다음 섹터의 첫 바이트를 읽어야 하므로 이를 처리하라는 말
			/* This cluster access spans a sector boundary in the FAT	*/
			/* There are a number of strategies to handling this. The 	*/
			/* easiest is to always load FAT sectors into memory 	*/
			/* in pairs if the volume is FAT12 (if you want to load	*/
			/* FAT sector N, you also load FAT sector N+1 immediately	*/
			/* following it in memory unless sector N is the last FAT	*/
			/* sector). It is assumed that this is the strategy used here	*/
			/* which makes this if test for a sector boundary span	*/
			/* unnecessary.						*/
			fs->sio.write( &fs->sio, ThisFATSecNum , ThisFATEntOffset, &fs->fat[ThisFATEntOffset], 1 );
			fs->sio.write( &fs->sio, ThisFATSecNum + 1, 0, &fs->fat[ThisFATEntOffset+1], 1 );
		}
		else
		{
			fs->sio.write( &fs->sio, ThisFATSecNum , ThisFATEntOffset, &fs->fat[ThisFATEntOffset], 2 );
		}
	}
	else
	if( fs->type == FS_FAT16 )
	{
		// untested
		int FATOffset = cluster + (cluster/2);

		int ThisFATSecNum = bpb->BPB_RsvdSecCnt + ( FATOffset / bytes_per_sec);
		int ThisFATEntOffset = (FATOffset % bytes_per_sec);

		int16u* fat16 = (int16u*)fs->fat;
		// update fat memory
		fat16[cluster] = value;

		// update fat real sector
		fs->sio.write( &fs->sio, ThisFATSecNum , ThisFATEntOffset, &fs->fat[ThisFATEntOffset], 2 );
	}

	/*
	{
	char* cur;
	int i;
	// update real fat
	cur = (char*)fs->fat;
	for( i = 0; i < fs->bpb.BPB_FATSz16 ; i ++ )
	{
		fs->sio.write( &fs->sio, i + fs->bpb.BPB_RsvdSecCnt, 0, cur, fs->bpb.BPB_BytesPerSec );
		cur += fs->bpb.BPB_BytesPerSec;
	}
	}
	*/
}


int fatfs_alloc_cluster( fatfs* fs )
{
	int first_sector;
	int cluster;
		
	cluster = fatfs_find_free_cluster( fs );
	if( cluster == NOT_FOUND_CLUSTER )
		return 0;

	// mark on FAT (cluster table)
	fatfs_set_fat( fs, cluster, 0xFFFFFFFF );

	// empty allocated cluster
	first_sector = get_first_sector_of_cluster( &fs->bpb, cluster );

	
	{
		int i;
		char empty_buffer[512*2];
		memset( empty_buffer, 0, sizeof(empty_buffer) );
		// fat cleanning (todo: tunning )
		for( i = 0 ; i < fs->bpb.BPB_SecPerClus; i ++ )
		{
			fs->sio.write( &fs->sio, first_sector+i, 0, empty_buffer, fs->sio.bytes_per_sector );
		}
	}
	
	return cluster;
}

int fatfs_realloc_cluster( fatfs* fs, int cluster )
{
	int new_cluster;
	int last_cluster;

	new_cluster = fatfs_alloc_cluster( fs );
	last_cluster = fatfs_get_last_cluster( fs, cluster );

	// append new cluster
	fatfs_set_fat( fs, last_cluster, new_cluster );
	return new_cluster;
}

int fatfs_create_directory( fatfs* fs, char* path )
{
	char parent[MAX_FSPATH], create_dir[MAX_FSPATH];
	char* parent_end;
	int cluster, next_cluster;
	fatfs_path p;
	directory_entry* de;
	
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);
	
	parent_end = get_parent_path( path, parent );
	kernel_printf("parent_end : %s \n", parent_end);
	if( parent_end == 0 )
	{
		// root path can't create
		return 0;
	}

	strcpy( create_dir, parent_end );
	kernel_printf("create_dir : %s \n", create_dir);

	if( fatfs_open_path( fs, parent, &p ) == 0 )
	{
		// no parent path
		return 0;
	}

	cluster = p.cluster;

	while( cluster != END_OF_CLUSTER )
	{
		char* limit;

		// root cluster isn't equal general cluster limit
		if( cluster == ROOT_CLUSTER && (fs->type == FS_FAT12 || fs->type == FS_FAT16) )
			limit = (p.buf+FATFS_MAX_PATH_BUF_SECTOR*bytes_per_sec);
		else
			limit = (p.buf+fs->bpb.BPB_SecPerClus*bytes_per_sec);
		
		// in this cluster
		de = (directory_entry*)p.buf;
		if( !de ) break;

		while( 1 )
		{
			if( (char*)de >= limit )
				break;	

			if( ( ((char*)de-(char*)p.buf > 2) && 
				(de-2)->DIR_FileName[0] == 0xE5 && 
				(de-1)->DIR_FileName[0] == 0xE5 && 
				de->DIR_FileName[0] == 0xE5 ) || // unused
				de->DIR_FileName[0] == 0x00 ) // end of cluster
			{
				char init_de[512*4];
				int first_sector;
				int alloc_cluster;
				directory_entry* pde;

				pde = de;
				memset( pde, 0, sizeof(directory_entry)*2 ); 

				pde->DIR_FileName[0] = 0xE5;
				pde++;

				// set DIR_FileName
				fatfs_set_file_full_name( pde->DIR_FileName, create_dir );
				kernel_printf("create_dir : %s \n", create_dir);
				kernel_printf("pde->DIR_FileName : %s \n", pde->DIR_FileName);
				pde->DIR_NTRes = 0x08;
				pde->DIR_Attr = FILE_ATTR_DIRECTORY;

				// create sub folder (default . / .. )
				alloc_cluster = fatfs_alloc_cluster( fs );
				if( alloc_cluster == 0 )
					return 0;

				pde->DIR_FstClusLO = alloc_cluster;

				{
					int sector = ( (int8u*)de - p.buf ) / bytes_per_sec;
					int offset = ( (int8u*)de - p.buf ) % bytes_per_sec;
					
					first_sector = get_first_sector_of_cluster( &fs->bpb, p.cluster ); 
					sector += first_sector;
					fs->sio.write( &fs->sio, sector, offset, (int8u*)de, sizeof(directory_entry)*2 );
				}
				
				// set subfolder's default directory entries ( . / .. )
				memset( init_de, 0, sizeof(init_de) );
				pde = (directory_entry*)init_de;
				strcpy( pde->DIR_FileName, ".          " );
				kernel_printf("pde->DIR_FileName : %s \n", pde->DIR_FileName);
				//fatfs_set_file_full_name( pde->DIR_FileName, "." );
				pde->DIR_Attr = FILE_ATTR_DIRECTORY;
				pde->DIR_FstClusLO = alloc_cluster;
				//C0 BB 04 31 04 31
				pde->DIR_CrtTime = 0xBBC0;
				pde->DIR_CrtDate = 0x3104;
				pde->DIR_LstAccDate = 0x3104;
				//C1 BB 04 31
				pde->DIR_WrtTime = 0xBBC1;
				pde->DIR_WrtDate = 0x3104;

				pde++;
				//fatfs_set_file_full_name( pde->DIR_FileName, ".." );
				strcpy( pde->DIR_FileName, "..         " );
				pde->DIR_Attr = FILE_ATTR_DIRECTORY;
				pde->DIR_FstClusLO = cluster;
				//C0 BB 04 31 04 31
				pde->DIR_CrtTime = 0xBBC0;
				pde->DIR_CrtDate = 0x3104;
				pde->DIR_LstAccDate = 0x3104;
				//C1 BB 04 31
				pde->DIR_WrtTime = 0xBBC1;
				pde->DIR_WrtDate = 0x3104;

				first_sector = get_first_sector_of_cluster( &fs->bpb, alloc_cluster );
				fs->sio.write( &fs->sio, first_sector, 0, init_de, sizeof(directory_entry)*2 );

				return 1;
			}

			de++;
		}

		// next cluster
		next_cluster = fatfs_read_next_cluster( fs, &p );
		
		// alloc new cluster
		if( next_cluster == END_OF_CLUSTER )
		{
			// alloc new cluster
			next_cluster = fatfs_realloc_cluster( fs, cluster );			
			
			fatfs_read_work_cluster( fs, &p, next_cluster );
		}
		cluster = next_cluster;
	}	

	// full disk
	return 0;
}

int fatfs_delete_directory( fatfs* fs, char* path )
{
	char parent[MAX_FSPATH], dir[MAX_FSPATH];
	char* parent_end;
	int cluster;
	fatfs_path p;
	directory_entry* de;
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);

	parent_end = get_parent_path( path, parent );
	if( parent_end == 0 )
	{
		// root path can't create
		return 0;
	}

	strcpy( dir, parent_end );

	if( fatfs_open_path( fs, parent, &p ) == 0 )
	{
		// no parent path
		return 0;
	}

	// in this cluster
	cluster = p.cluster;

	while( cluster != END_OF_CLUSTER )
	{
		// in this cluster
		de = (directory_entry*)p.buf;

		while( de && (int8u*)de < (p.buf+bytes_per_sec) )
		{
			if( fatfs_is_equal_fname( de->DIR_FileName, dir ) )
			{
				// now delete this dir
				de->DIR_FileName[0] = 0xE5; // set unused
				
				// update sector
				{
					int first_sector;
					int sector = ( (int8u*)de - p.buf ) / bytes_per_sec;
					int offset = ( (int8u*)de - p.buf ) % bytes_per_sec;
					
					first_sector = get_first_sector_of_cluster( &fs->bpb, p.cluster ); 
					sector += first_sector;
					fs->sio.write( &fs->sio, sector, offset, (int8u*)de, sizeof(directory_entry)*1 );
				}

				// update fat table
				fatfs_free_cluster_list( fs, de->DIR_FstClusLO );
				// success
				return 1;
			}
			
			// next directory entry
			de++;
		}
		// next cluster
		cluster = fatfs_read_next_cluster( fs, &p );
	}	

	// full disk
	return 0;
}

int fatfs_set_unused_de( fatfs* fs, char* path, directory_entry* pde )
{
	char parent[MAX_FSPATH], dir[MAX_FSPATH];
	char* parent_end;
	int cluster;
	fatfs_path p;
	directory_entry* de;
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);

	parent_end = get_parent_path( path, parent );
	if( parent_end == 0 )
	{
		// root path can't create
		return 0;
	}

	strcpy( dir, parent_end );

	if( fatfs_open_path( fs, parent, &p ) == 0 )
	{
		// no parent path
		return 0;
	}

	// in this cluster
	cluster = p.cluster;

	while( cluster != END_OF_CLUSTER )
	{
		// in this cluster
		de = (directory_entry*)p.buf;

		while( de && (int8u*)de < (p.buf+bytes_per_sec) )
		{
			if( fatfs_is_equal_fname( de->DIR_FileName, dir ) )
			{
				// save directory entry
				memcpy( pde, de, sizeof(directory_entry) );

				// now delete this dir
				de->DIR_FileName[0] = 0xE5; // set unused

				// update sector
				{
					int first_sector;
					int sector = ( (int8u*)de - p.buf ) / bytes_per_sec;
					int offset = ( (int8u*)de - p.buf ) % bytes_per_sec;
					
					first_sector = get_first_sector_of_cluster( &fs->bpb, p.cluster ); 
					sector += first_sector;
					fs->sio.write( &fs->sio, sector, offset, (int8u*)de, sizeof(directory_entry)*1 );
				}

				return de->DIR_FstClusLO;
			}

			// next directory entry
			de++;
		}
		// next cluster
		cluster = fatfs_read_next_cluster( fs, &p );
	}	

	// full disk
	return 0;
}


int fatfs_create_de( fatfs* fs, char* path, directory_entry* psde )
{
	char parent[MAX_FSPATH], create_dir[MAX_FSPATH];
	char* parent_end;
	int cluster, next_cluster;
	fatfs_path p;
	directory_entry* de;
	
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);

	parent_end = get_parent_path( path, parent );
	if( parent_end == 0 )
	{
		// root path can't create
		return 0;
	}

	strcpy( create_dir, parent_end );

	// parent => path가 '/' 으로 나온다.
	if( fatfs_open_path( fs, parent, &p ) == 0 )
	{
		// no parent path
		return 0;
	}

	cluster = p.cluster;

	while( cluster != END_OF_CLUSTER )
	{
		// in this cluster
		de = (directory_entry*)p.buf;

		while( de && (int8u*)de < (p.buf+bytes_per_sec) )
		{
			if( ( ((char*)de-(char*)p.buf > 2) && 
				(de-2)->DIR_FileName[0] == 0xE5 && 
				(de-1)->DIR_FileName[0] == 0xE5 && 
				de->DIR_FileName[0] == 0xE5 ) || // unused
				de->DIR_FileName[0] == 0x00 ) // end of cluster
			{
				int first_sector;
				//int alloc_cluster;
				directory_entry* pde;

				pde = de;
				memset( pde, 0, sizeof(directory_entry)*2 ); 

				pde->DIR_FileName[0] = 0xE5;
				pde++;

				memcpy( pde, psde, sizeof(directory_entry) );
				fatfs_set_file_full_name( pde->DIR_FileName, create_dir );

				{
					int sector = ( (int8u*)de - p.buf ) / bytes_per_sec;
					int offset = ( (int8u*)de - p.buf ) % bytes_per_sec;
					
					first_sector = get_first_sector_of_cluster( &fs->bpb, p.cluster ); 
					sector += first_sector;
					fs->sio.write( &fs->sio, sector, offset, (int8u*)de, sizeof(directory_entry)*2 );
				}

				return 1;
			}

			de++;
		}

		// next cluster
		next_cluster = fatfs_read_next_cluster( fs, &p );
		
		// alloc new cluster
		if( next_cluster == END_OF_CLUSTER )
		{
			// alloc new cluster
			next_cluster = fatfs_realloc_cluster( fs, cluster );			
			
			fatfs_read_work_cluster( fs, &p, next_cluster );
		}
		cluster = next_cluster;
	}	

	// full disk
	return 0;
}

int fatfs_rename_directory( fatfs* fs, char* path, char* newpath )
{
	/*
	de->DIR_LoCluster인 data cluster를 기억한 후
	자신의 DIR_FileName은 지운 후, 새로 붙여야 할 곳의
	빈곳에 de를 놓은 후 data cluster를 연결하자.
  */
	// 새로운 곳에 마련한후 연결할 데이타 클러스터
	directory_entry de;
	
	if( !fatfs_set_unused_de( fs, path, &de ) )
		return 0;
	
	return fatfs_create_de( fs, newpath, &de );
}

int fatfs_create_file( fatfs* fs, char* path )
{
	char parent[MAX_FSPATH], create_name[MAX_FSPATH];
	char* parent_end;
	int cluster, next_cluster;
	fatfs_path p;
	directory_entry* de;
	int bytes_per_sec = fs->bpb.BPB_BytesPerSec[0] +(fs->bpb.BPB_BytesPerSec[1] << 8);

	parent_end = get_parent_path( path, parent );
	if( parent_end == 0 )
	{
		// root path can't create
		return 0;
	}

	strcpy( create_name, parent_end );

	if( fatfs_open_path( fs, parent, &p ) == 0 )
	{
		// no parent path
		return 0;
	}

	// in this cluster
	cluster = p.cluster;

	while( cluster != END_OF_CLUSTER )
	{
		// in this cluster
		de = (directory_entry*)p.buf;

		while( de && (int8u*)de < (p.buf+bytes_per_sec) )
		{
			if( ( ((char*)de-(char*)p.buf > 2) && 
				(de-2)->DIR_FileName[0] == 0xE5 && 
				(de-1)->DIR_FileName[0] == 0xE5 && 
				de->DIR_FileName[0] == 0xE5 ) || // unused
				de->DIR_FileName[0] == 0x00 ) // end of cluster
			{
				int first_sector;
				int alloc_cluster;
				directory_entry* pde;

				pde = de;
				memset( pde, 0, sizeof(directory_entry)*2 ); 

				pde->DIR_FileName[0] = 0xE5;
				pde++;

				// set DIR_FileName
				fatfs_set_file_full_name( pde->DIR_FileName, create_name );
				pde->DIR_NTRes = 0x18;
				pde->DIR_Attr = 0x00;
				pde->DIR_FileSize = 0;

				// create file data cluster
				alloc_cluster = fatfs_alloc_cluster( fs );
				if( alloc_cluster == 0 )
				{
					return 0;
				}

				pde->DIR_FstClusLO = alloc_cluster;

				{
					int sector = ( (int8u*)de - p.buf ) / bytes_per_sec;
					int offset = ( (int8u*)de - p.buf ) % bytes_per_sec;
					
					first_sector = get_first_sector_of_cluster( &fs->bpb, p.cluster ); 
					sector += first_sector;
					fs->sio.write( &fs->sio, sector, offset, (int8u*)de, sizeof(directory_entry)*2 );
				}	

				return 1;
			}

			de++;
		}

		// next cluster
		next_cluster = fatfs_read_next_cluster( fs, &p );
		
		// alloc new cluster
		if( next_cluster == END_OF_CLUSTER )
		{
			// alloc new cluster
			next_cluster = fatfs_realloc_cluster( fs, cluster );			
			
			fatfs_read_work_cluster( fs, &p, next_cluster );
		}
		cluster = next_cluster;
	}	

	// full disk
	return 0;
	
}

int fatfs_rename_file( fatfs* fs, char* path, char* newpath )
{
	return fatfs_rename_directory( fs, path, newpath );
}

int fatfs_delete_file( fatfs* fs, char* path )
{
	return fatfs_delete_directory( fs, path );
}
